from setuptools import setup, Extension, find_packages, find_namespace_packages
from setuptools.command.build_ext import build_ext
import sys
import setuptools
from glob import glob
import distutils.ccompiler

# Bump the major version when making backwards incompatible changes.
MAJOR = 0

# Bump the minor version when adding backwards compatible features.
MINOR = 0

# Bump the patch version when making bug fixes.
PATCH = 12

__version__ = f'{MAJOR}.{MINOR}.{PATCH}'

# Note the nonstandard file configuration in this setup.py.  In the main BOOM
# repository stored on github, setup.py and the pybind11 bindings are kept in
# .../Interfaces/python/BayesBoom/... For setup.py to work the C++ code must be
# in a subdirectory below setup.py.
#
# A script in the top level of the BOOM project copies the BOOM source code
# into a build directory in a way that will make setup.py happy.  This file is
# intended to be run by that build script, and not directly from the
# repository.


class get_pybind_include(object):
    """Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)


BOOM_DIR = "BayesBoom/boom"
BOOM_DIR += "/"

# We keep track of the boom headers here just in case setup.py support for
# header files improves one day.  The MANIFEST.in file instructs the package to
# include them.
boom_headers = glob(BOOM_DIR + "*.hpp")

eigen_headers = glob(BOOM_DIR + "Eigen")
boom_headers += eigen_headers

distributions_sources = glob(BOOM_DIR + "distributions/*.cpp")
distributions_headers = (
    [BOOM_DIR + "distributions.hpp"]
    + glob(BOOM_DIR + "distributions/*.hpp")
    )
boom_headers += distributions_headers

linalg_sources = glob(BOOM_DIR + "LinAlg/*.cpp")
linalg_headers = glob(BOOM_DIR + "LinAlg/*.hpp")
boom_headers += linalg_headers

math_sources = glob(BOOM_DIR + "math/*.cpp")
math_sources += glob(BOOM_DIR + "math/cephes/*.cpp")

numopt_sources = glob(BOOM_DIR + "numopt/*.cpp")
numopt_headers = [BOOM_DIR + "numopt.hpp"] + glob(BOOM_DIR + "numopt/*.hpp")
boom_headers += numopt_headers

rmath_sources = glob(BOOM_DIR + "Bmath/*.cpp")
rmath_headers = glob(BOOM_DIR + "Bmath/*.hpp")
boom_headers += rmath_headers

rmath_sources = glob(BOOM_DIR + "Bmath/*.cpp")
rmath_headers = glob(BOOM_DIR + "Bmath/*.hpp")

samplers_sources = glob(BOOM_DIR + "Samplers/*.cpp")
samplers_sources += [BOOM_DIR + "Samplers/Gilks/arms.cpp"]
samplers_headers = glob(BOOM_DIR + "Samplers/*.hpp")
boom_headers += samplers_headers

stats_sources = glob(BOOM_DIR + "stats/*.cpp")
stats_headers = glob(BOOM_DIR + "stats/*.hpp")
boom_headers += stats_headers

targetfun_sources = glob(BOOM_DIR + "TargetFun/*.cpp")
targetfun_headers = glob(BOOM_DIR + "TargetFun/*.hpp")
boom_headers += targetfun_headers

utils_sources = glob(BOOM_DIR + "cpputil/*.cpp")
utils_headers = glob(BOOM_DIR + "cpputil/*.hpp")
boom_headers += utils_headers

models_sources = (
    glob(BOOM_DIR + "Models/*.cpp")
    + glob(BOOM_DIR + "Models/PosteriorSamplers/*.cpp")
    + glob(BOOM_DIR + "Models/Policies/*.cpp"))
models_headers = (
    glob(BOOM_DIR + "Models/*.hpp")
    + glob(BOOM_DIR + "Models/Policies/*.hpp")
    + glob(BOOM_DIR + "Models/PosteriorSamplers/*.hpp"))
boom_headers += models_headers

# Specific model classes to be added later, glm's hmm's, etc.
bart_sources = (
    glob(BOOM_DIR + "Models/Bart/*.cpp")
    + glob(BOOM_DIR + "Models/Bart/PosteriorSamplers/*.cpp")
    )
bart_headers = (
    glob(BOOM_DIR + "Models/Bart/*.hpp")
    + glob(BOOM_DIR + "Models/Bart/PosteriorSamplers/*.hpp")
    )
boom_headers += bart_headers

glm_sources = (
    glob(BOOM_DIR + "Models/Glm/*.cpp")
    + glob(BOOM_DIR + "Models/Glm/PosteriorSamplers/*.cpp")
    )
glm_headers = (
    glob(BOOM_DIR + "Models/Glm/*.hpp")
    + glob(BOOM_DIR + "Models/Glm/PosteriorSamplers/*.hpp")
    )
boom_headers += glm_headers

hmm_sources = (
    glob(BOOM_DIR + "Models/HMM/*.cpp")
    + glob(BOOM_DIR + "Models/HMM/Clickstream/*.cpp")
    + glob(BOOM_DIR + "Models/HMM/Clickstream/PosteriorSamplers/*.cpp")
    + glob(BOOM_DIR + "Models/HMM/PosteriorSamplers/*.cpp")
    )
hmm_headers = (
    glob(BOOM_DIR + "Models/HMM/*.hpp")
    + glob(BOOM_DIR + "Models/HMM/Clickstream/*.hpp")
    + glob(BOOM_DIR + "Models/HMM/Clickstream/PosteriorSamplers/*.hpp")
    + glob(BOOM_DIR + "Models/HMM/PosteriorSamplers/*.hpp")
    )
boom_headers += hmm_headers

hierarchical_sources = (
    glob(BOOM_DIR + "Models/Hierarchical/*.cpp")
    + glob(BOOM_DIR + "Models/Hierarchical/PosteriorSamplers/*.cpp")
    )
hierarchical_headers = (
    glob(BOOM_DIR + "Models/Hierarchical/*.hpp")
    + glob(BOOM_DIR + "Models/Hierarchical/PosteriorSamplers/*.hpp")
    )
boom_headers += hierarchical_headers

impute_sources = (
    glob(BOOM_DIR + "Models/Impute/*.cpp")
    )
impute_headers = (
    glob(BOOM_DIR + "Models/Impute/*.hpp")
    )
boom_headers += impute_headers

irt_sources = (
    glob(BOOM_DIR + "Models/IRT/*.cpp")
    + glob(BOOM_DIR + "Models/IRT/PosteriorSamplers/*.cpp")
    )
irt_headers = (
    glob(BOOM_DIR + "Models/IRT/*.hpp")
    + glob(BOOM_DIR + "Models/IRT/PosteriorSamplers/*.hpp")
    )
boom_headers += irt_headers

mixture_sources = (
    glob(BOOM_DIR + "Models/Mixtures/*.cpp")
    + glob(BOOM_DIR + "Models/Mixtures/PosteriorSamplers/*.cpp")
    )
mixture_headers = (
    glob(BOOM_DIR + "Models/Mixtures/*.hpp")
    + glob(BOOM_DIR + "Models/Mixtures/PosteriorSamplers/*.hpp")
    )
boom_headers += mixture_headers

nnet_sources = (
    glob(BOOM_DIR + "Models/Nnet/*.cpp")
    + glob(BOOM_DIR + "Models/Nnet/PosteriorSamplers/*.cpp")
    )
nnet_headers = (
    glob(BOOM_DIR + "Models/Nnet/*.hpp")
    + glob(BOOM_DIR + "Models/Nnet/PosteriorSamplers/*.hpp")
    )
boom_headers += nnet_headers

point_process_sources = (
    glob(BOOM_DIR + "Models/PointProcess/*.cpp")
    + glob(BOOM_DIR + "Models/PointProcess/PosteriorSamplers/*.cpp")
    )
point_process_headers = (
    glob(BOOM_DIR + "Models/PointProcess/*.hpp")
    + glob(BOOM_DIR + "Models/PointProcess/PosteriorSamplers/*.hpp")
    )
boom_headers += point_process_headers

state_space_sources = (
    glob(BOOM_DIR + "Models/StateSpace/*.cpp")
    + glob(BOOM_DIR + "Models/StateSpace/Filters/*.cpp")
    + glob(BOOM_DIR + "Models/StateSpace/PosteriorSamplers/*.cpp")
    + glob(BOOM_DIR + "Models/StateSpace/StateModels/*.cpp")
    + glob(BOOM_DIR + "Models/StateSpace/StateModels/PosteriorSamplers/*.cpp")
)
state_space_headers = (
    glob(BOOM_DIR + "Models/StateSpace/*.hpp")
    + glob(BOOM_DIR + "Models/StateSpace/Filters/*.hpp")
    + glob(BOOM_DIR + "Models/StateSpace/PosteriorSamplers/*.hpp")
    + glob(BOOM_DIR + "Models/StateSpace/StateModels/PosteriorSamplers/*.hpp")
)
boom_headers += state_space_headers

time_series_sources = (
    glob(BOOM_DIR + "Models/TimeSeries/*.cpp")
    + glob(BOOM_DIR + "Models/TimeSeries/PosteriorSamplers/*.cpp")
    )
time_series_headers = (
    glob(BOOM_DIR + "Models/TimeSeries/*.hpp")
    + glob(BOOM_DIR + "Models/TimeSeries/PosteriorSamplers/*.hpp")
    )
boom_headers += time_series_headers

test_utils_sources = glob(BOOM_DIR + "test_utils/*.cpp")


boom_library_sources = (
    distributions_sources
    + linalg_sources
    + math_sources
    + numopt_sources
    + rmath_sources
    + samplers_sources
    + stats_sources
    + targetfun_sources
    + utils_sources
    + models_sources
    + bart_sources
    + glm_sources
    + hmm_sources
    + hierarchical_sources
    + impute_sources
    + irt_sources
    + mixture_sources
    + nnet_sources
    + point_process_sources
    + state_space_sources
    + time_series_sources
    + test_utils_sources
)

boom_extension_sources = (
    [BOOM_DIR + "pybind11/module.cpp"]
    + glob(BOOM_DIR + "pybind11/Models/*.cpp")
    + glob(BOOM_DIR + "pybind11/Models/Glm/*.cpp")
    + glob(BOOM_DIR + "pybind11/Models/Impute/*.cpp")
    + glob(BOOM_DIR + "pybind11/Models/StateSpace/*.cpp")
    + glob(BOOM_DIR + "pybind11/Models/StateSpace/StateModels/*.cpp")
    + glob(BOOM_DIR + "pybind11/Models/TimeSeries/*.cpp")
    + glob(BOOM_DIR + "pybind11/LinAlg/*.cpp")
    + glob(BOOM_DIR + "pybind11/stats/*.cpp")
    + glob(BOOM_DIR + "pybind11/cpputil/*.cpp")
    + glob(BOOM_DIR + "pybind11/distributions/*.cpp")
    + glob(BOOM_DIR + "pybind11/test_utils/*.cpp")
)

boom_sources = boom_extension_sources + boom_library_sources


# ---------------------------------------------------------------------------
# From
# https://stackoverflow.com/questions/11013851/speeding-up-build-process-with-distutils
# monkey-patch for parallel compilation
def parallelCCompile(self, sources, output_dir=None, macros=None,
                     include_dirs=None, debug=0, extra_preargs=None,
                     extra_postargs=None, depends=None):

    # those lines are copied from distutils.ccompiler.CCompiler directly
    macros, objects, extra_postargs, pp_opts, build = self._setup_compile(
        output_dir, macros, include_dirs, sources, depends, extra_postargs)

    cc_args = self._get_cc_args(pp_opts, debug, extra_preargs)
    # parallel code
    nthreads = 16  # number of parallel compilations
    import multiprocessing.pool

    def _single_compile(obj):
        try:
            src, ext = build[obj]
        except KeyError:
            return
        self._compile(obj, src, ext, cc_args, extra_postargs, pp_opts)

    # convert to list, imap is evaluated on-demand
    multiprocessing.log_to_stderr()
    with multiprocessing.pool.ThreadPool(nthreads) as pool:
        it = pool.imap(_single_compile, objects)
        list(it)

    # list(multiprocessing.pool.ThreadPool(nthreads).imap(
    #     _single_compile, objects))

    return objects


distutils.ccompiler.CCompiler.compile = parallelCCompile
# End of parallel compile "monkey patch"
# ---------------------------------------------------------------------------

# remove trailing slash from BOOM_DIR
if BOOM_DIR.endswith("/"):
    BOOM_DIR = BOOM_DIR[:-1]

ext_modules = [
    Extension(
        '_boom',
        sources=boom_sources,
        depends=boom_headers,
        headers=boom_headers,
        include_dirs=[
            "./BayesBoom/boom",
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True)
        ],
        language='c++'
    ),
]


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14/17] compiler flag.

    The newer version is prefered over c++11 (when it is available).
    """
    flags = ['-std=c++17', '-std=c++14', '-std=c++11']

    for flag in flags:
        if has_flag(compiler, flag):
            return flag

    raise RuntimeError('Unsupported compiler -- at least C++11 support '
                       'is needed!')


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }
    l_opts = {
        'msvc': [],
        'unix': [],
    }

    if sys.platform == 'darwin':
        darwin_opts = ['-stdlib=libc++',
                       '-mmacosx-version-min=10.14',
                       '-Wno-sign-compare']
        c_opts['unix'] += darwin_opts
        l_opts['unix'] += darwin_opts
    elif sys.platform == 'linux':
        c_opts['unix'] = ['-Wno-sign-compare']

    def build_extensions(self):
        ct = self.compiler.compiler_type
        print(f"compiler type is: {ct}")
        opts = self.c_opts.get(ct, [])
        link_opts = self.l_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())  # noqa
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')

            # For deubgging purposes only.  Do not submit code with this option
            # present.
            # opts.append("-O0")
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())  # noqa
        for ext in self.extensions:
            ext.extra_compile_args = opts
            ext.extra_link_args = link_opts
        build_ext.build_extensions(self)


def FindPackagesAndBlab():
    """
    Find the sub-packages to be installed, and blab about them in print
    statements during the build process.
    """
    # packages = ["R", "bsts", "spikeslab", "dynreg", "test_utils",
    # "boom/pybind11"]
    packages = find_namespace_packages(include=["BayesBoom.*"],
                                       exclude=["BayesBoom.*.*"])
    # packages = find_packages()
    if len(packages) == 0:
        packages = find_packages()
    print(f"***** HEY!! I found the following packages: {packages} *****")
    print(f"      The BOOM_DIR argument from setup.py is {BOOM_DIR}")
    if len(packages) == 0:
        raise Exception("No packages found.")

    return packages


setup(
    name='BayesBoom',
    packages=FindPackagesAndBlab(),
    version=__version__,
    author='Steven L. Scott',
    author_email='steve.the.bayesian@gmail.com',
    url='https://github.com/steve-the-bayesian/BOOM',
    description='Tools for Bayesian modeling.',
    long_description="""Boom stands for 'Bayesian object oriented modeling'.
    It is also the sound your computer makes when it crashes.

    The main part of the Boom library is formulated in terms of abstractions
    for Model, Data, Params, and PosteriorSampler.  A Model is primarily an
    environment where parameters can be learned from data.  The primary
    learning method is Markov chain Monte Carlo, with custom samplers defined
    for specific models.

    The archetypal Boom program looks something like this:

    import BayesBoom as Boom

    some_data = 3 * np.random.randn(100) + 7
    model = Boom.GaussianModel()
    model.set_data(some_data)
    precision_prior = Boom.GammaModel(0.5, 1.5)
    mean_prior = Boom.GaussianModel(0, 10**2)
    poseterior_sampler = Boom.GaussianSemiconjugateSampler(
        model, mean_prior, precision_prior)
    model.set_method(poseterior_sampler)
    niter = 100
    mean_draws = np.zeros(niter)
    sd_draws = np.zeros(niter)
    for i in range(100):
        model.sample_posterior()
        mean_draws[i] = model.mu()
        sd_draws[i] = model.sigma()

    """,
    long_description_content_type="text/plain",
    ext_modules=ext_modules,
    install_requires=['pybind11>=2.3'],
    setup_requires=['pybind11>=2.3'],
    cmdclass={'build_ext': BuildExt},
    zip_safe=False,
)



import BayesBoom.boom as boom
import pandas as pd
from pandas.api.types import (
    is_numeric_dtype, is_categorical_dtype, is_object_dtype, is_bool_dtype
)


def to_data_table(data: pd.DataFrame):
    """
    Create a BOOM DataTable object from a pandas DataFrame.  The categories of
    any categorical variables will be handled as strings.
    """
    dtypes = data.dtypes
    ans = boom.DataTable()
    for i in range(data.shape[1]):
        dt = dtypes[i]
        vname = data.columns[i]
        if is_numeric_dtype(dt) or is_bool_dtype(dt):
            ans.add_numeric(boom.Vector(data.iloc[:, i].values.astype("float")),
                            vname)
        elif is_categorical_dtype(dt):
            x = data.iloc[:, i]
            values = x.cat.codes
            codes = x.cat.categories
            ans.add_categorical(values, codes, vname)
        elif is_object_dtype(dt):
            labels = data.iloc[:, i].astype("str")
            ans.add_categorical_from_labels(labels.values, vname)
        else:
            raise Exception(
                f"Only numeric or categorical data are supported.  "
                f"Column {i} ({data.columns[i]}) has dtype {dt}."
            )
    return ans


def to_data_frame(data: boom.DataTable, columns=None, index=None):
    """
    Convert a boom.DataTable to a pd.DataFrame.

    Args:
      data:  The data table to be converted.
      columns: The column names.  If None then the names of the data table will
        be used.
      index:  The index to be applied to the returned data frame.

    Returns:
      A pandas data frame containing the data from 'data'.
    """
    if columns is None:
        columns = data.variable_names

    if len(columns) == 0:
        columns = ["V" + str(i) for i in range(data.ncol)]

    if len(columns) != data.ncol:
        raise Exception("The number of entries in 'columns' must match "
                        "the number of variables in the data table.")

    if index is None:
        index = range(data.nrow)
    if len(index) != data.nrow:
        raise Exception("The number of entries in 'index' must match "
                        "the number of rows in the data table.")

    ans_as_dict = {}
    for i, vname in enumerate(columns):
        vtype = data.variable_type(i)
        if vtype == "numeric":
            ans_as_dict[vname] = data.getvar(i).to_numpy()
        elif vtype == "categorical":
            values = data.get_nominal_values(i)
            levels = data.get_nominal_levels(i)
            ans_as_dict[vname] = pd.Categorical.from_codes(values, levels)
        else:
            raise Exception("Only numeric or categorical values are supported.")

    return pd.DataFrame(ans_as_dict, index=index)



from .bsts import ObservationModelManager
import numpy as np
import pandas as pd
import patsy
import scipy
import BayesBoom.boom as boom
import BayesBoom.R as R
import BayesBoom.spikeslab as spikeslab


class GaussianStateSpaceModelFactory:
    """
    Model factory for models with Gassian observation errors.  Thus the handled
    models could either be boom.StateSpaceModel or
    boom.StateSpaceRegressionModel.
    """
    def __init__(self):
        self._model = None
        self.predictor_names = None

    def create_model(self, prior: R.SdPrior, data: pd.Series, rng):
        """
        Args:
          prior: an R.SdPrior object describing the prior distribution on the
            residual variance paramter.
          data:  The time series of observations as a Pandas Series.
          rng: The boom random number generator.

        Returns:
          A boom.StateSpaceModel object.
        """
        if data is not None:
            if (isinstance(data, np.ndarray)):
                boom_data = boom.Vector(data.ravel().astype(float))
                is_observed = np.isfinite(data)
            else:
                boom_data = boom.Vector(data.values.astype(float))
                is_observed = np.isfinite(data)
            self._model = boom.StateSpaceModel(boom_data, is_observed)
        else:
            self._model = boom.StateSpaceModel()

        if prior is None:
            if data is None:
                raise Exception("One of 'prior' or 'data' must be given")
            sdy = np.std(data, ddof=1)
            prior = R.SdPrior(sigma_guess=sdy, upper_limit=sdy * 1.2)

        if not isinstance(prior, R.SdPrior):
            raise Exception(
                "I expected a prior of type R.SdPrior on the residual variance"
                " parameter."
            )
        self._prior = prior

        boom_prior = boom.ChisqModel(prior.sample_size, prior.sigma_guess)
        observation_model_sampler = boom.ZeroMeanGaussianConjSampler(
            self._model.observation_model,
            boom_prior)
        observation_model_sampler.set_sigma_upper_limit(
            prior.upper_limit)
        self._model.observation_model.set_method(observation_model_sampler)

        sampler = boom.StateSpacePosteriorSampler(self._model, rng)
        self._model.set_method(sampler)

        if data is not None:
            self._original_series = data

        return self._model

    def create_observation_model_manager(self):
        return GaussianObservationModelManager(xdim=0, formula=None)


class StateSpaceRegressionModelFactory:

    def __init__(self, formula):
        if not isinstance(formula, str):
            raise Exception("formula must be a string.")
        self._formula = formula
        self.predictor_names = None

    def create_model(self, prior, data, rng, **kwargs):
        """
        Create the boom model object, and store related model artifacts.

        Args:
          formula:  A model formula describing the regression component.
          data: A pandas DataFrame containing the variables appearing
            'formula'.
          prior: A spikeslab.RegressionSpikeSlabPrior describing the prior
            distribution on the regression coefficients and the residual
            standard deviation.

        Effects: self._model is created, and model formula artifacts are stored
          so they will be available for future predictions.
        """
        response, predictors = patsy.dmatrices(self._formula, data)
        self.predictor_names = predictors.design_info.term_names
        boom_response = boom.Vector(response)
        boom_predictors = boom.Matrix(predictors)
        if prior is None:
            prior = spikeslab.RegressionSpikeSlabPrior(
                boom_predictors, boom_response, **kwargs)

        self._prior = prior

        if not isinstance(prior, spikeslab.RegressionSpikeSlabPrior):
            raise Exception("Unexpected type for prior.")
        is_observed = np.isfinite(response)
        self._model = boom.StateSpaceRegressionModel(
            boom_response, boom_predictors, is_observed)

        reg = self._model.observation_model
        observation_model_sampler = boom.BregVsSampler(
            reg,
            prior.slab(reg.Sigsq_prm),
            prior.residual_precision,
            prior.spike)
        reg.set_method(observation_model_sampler)
        self._original_series = response

        sampler = boom.StateSpacePosteriorSampler(
            self._model)
        self._model.set_method(sampler)

        return self._model

    def create_observation_model_manager(self):
        return GaussianObservationModelManager(self._model.xdim, self._formula)


class GaussianObservationModelManager(ObservationModelManager):
    """
    The observation model manager supports the observation model in the boom
    object. It holds posterior draws of observation model parameters, log
    likelihood, and metadata about the predictors, if any.
    """
    def __init__(self, xdim, formula):
        """
        Args:
          xdim: The dimension of the predictor variable.  For a pure time
            series model with no predictors, call with xdim == 0.
        """
        self._xdim = xdim
        self._formula = formula

    @property
    def has_regression(self):
        return self._xdim > 0

    def allocate_space(self, niter: int, time_dimension: int):
        """
        Create space to hold 'niter' MCMC draws.
        """
        self._residual_sd = np.empty(niter)
        self._log_likelihood = np.empty(niter)

        if self._xdim > 0:
            self._coefficients = scipy.sparse.lil_matrix((niter, self._xdim))

        if self.has_regression > 0:
            self._regression_contribution = np.empty((niter, time_dimension))

    def record_draw(self, iteration: int, model):
        """
        Record the current state of 'model' as of the specified 'iteration'.

        """
        self._residual_sd[iteration] = model.residual_sd
        self._log_likelihood[iteration] = model.log_likelihood
        if self._xdim > 0:
            self._coefficients[iteration, :] = spikeslab.sparsify(model.coef)
            self._regression_contribution[iteration, :] = (
                model.regression_contribution.to_numpy()
            )

    def restore_draw(self, iteration: int, model):
        """
        Set the state of 'model' to the requested iteration.
        """
        model.observation_model.set_sigma(self._residual_sd[iteration])
        if self._xdim > 0:
            model.observation_model.set_sigma(self._residual_sd[iteration])
            spikeslab.set_glm_coefs(model.observation_model.coefs,
                                    self._coefficients[iteration, :])

    def format_prediction_data(self, prediction_data, **kwargs):
        if self._xdim > 0:
            predictor_matrix = patsy.dmatrix(
                self._formula, data=prediction_data)
            formatted = {
                "forecast_horizon": prediction_data.shape[0],
                "predictors": boom.Matrix(predictor_matrix),
            }
        else:
            formatted = {
                "forecast_horizon": int(prediction_data)
            }
        return formatted

    def predict(self, model, formatted_prediction_data, boom_final_state, rng,
                separate_components=False, **kwargs):
        """
        Args:
          model: Either a boom.StateSpaceModel, or a
            boom.StateSpaceRegressionModel.
          formatted_prediction_data: The output of
            self.format_prediction_data().  A dict containing the prediction
            data in the form expected by the boom model.
          burn: The number of MCMC iterations to discard as burn-in.
          **kwargs: Unused.  Extra arguments will cause an exception to be
            raised.
        """
        extra_args = {**kwargs}
        if extra_args:
            raise Exception("Unexpected arguments were given to Bsts.predict:"
                            f"  {extra_args.keys()}.")

        if isinstance(model, boom.StateSpaceRegressionModel):
            if separate_components:
                draw = model.simulate_forecast_components(
                    rng,
                    formatted_prediction_data["predictors"],
                    boom_final_state)
            else:
                draw = model.simulate_forecast(
                    rng,
                    formatted_prediction_data["predictors"],
                    boom_final_state)
        elif isinstance(model, boom.StateSpaceModel):
            if separate_components:
                draw = model.simulate_forecast_components(
                    rng,
                    formatted_prediction_data["forecast_horizon"],
                    boom_final_state)
                # Insert a row of 0's for the regression component.
                draw = R.to_numpy(draw)
                draw = np.insert(draw, -1, np.zeros(draw.shape[1]), axis=0)
            else:
                draw = model.simulate_forecast(
                    rng,
                    formatted_prediction_data["forecast_horizon"],
                    boom_final_state)
        else:
            raise Exception(
                "Unrecognized boom model passed to "
                "GaussianObservationModelManager.predict: "
                f"{model.__class__.__name__}"
            )

        return R.to_numpy(draw)
